@@ -3,14 +3,14 @@ import re
 import json
 import logging
 import isodate
-from datetime import datetime, timedelta
-from typing import Union,List, Dict
+from datetime import datetime
+from typing import Union, List, Dict
 import http.client as http_client
 
 from urllib.parse import urlencode
 import requests
 from docassemble.base.functions import all_variables, get_config
-from docassemble.base.util import IndividualName, DAObject, log, Person, Individual, DADateTime, as_datetime, reconsider
+from docassemble.base.util import DAObject, log, Person, Individual, DADateTime, as_datetime, reconsider
 from docassemble.AssemblyLine.al_document import ALDocumentBundle
 from docassemble.AssemblyLine.al_general import ALIndividual
 from datetime import datetime
@@ -34,17 +34,34 @@ class ApiResponse(DAObject):
   def is_ok(self):
     return self.response_code in [200, 201, 202, 203, 204, 205]
 
-def _recursive_give_data_url(bundle):
-  """Prepares the document types by setting a semi-permanent enabled and a data url"""
+def _give_data_url(bundle: ALDocumentBundle):
+  """Prepares the filing documents by setting a semi-permanent enabled and a data url
+  The document bundle can either consist of documents or other document bundles. But each top element will
+  sent to Tyler as a separate filing document, and needs to have the necessary attributes set (tyler_filing_id, etc)
+  """
+  if bundle is None:
+    return
   for doc in bundle:
-    if isinstance(doc, ALDocumentBundle):
-      _recursive_give_data_url(doc)
-    else:
-      doc.proxy_enabled = doc.always_enabled or doc.enabled
-      if doc.proxy_enabled:
-        doc.data_url = doc.as_pdf().url_for(temporary=True)
-      if hasattr(doc, 'enabled'):
-        del doc.enabled
+    doc.proxy_enabled = doc.always_enabled or doc.enabled
+    if doc.proxy_enabled:
+      doc.data_url = doc.as_pdf().url_for(temporary=True)
+      doc.page_count = doc.as_pdf().num_pages()
+    if hasattr(doc, 'enabled'):
+      del doc.enabled
+
+def _get_all_vars(bundle: ALDocumentBundle):
+  """Strips out some extra big variables that we don't need to serialize and send across the network"""
+  _give_data_url(bundle)
+  all_vars_dict = all_variables()
+  all_vars_dict.pop('trial_court_resp', None)
+  all_vars_dict.pop('x', None)
+  all_vars_dict.pop('trial_court_options', None)
+  all_vars_dict.pop('found_case', None)
+  all_vars_dict.pop('selected_exisiting_case', None)
+  all_vars_dict.pop('filing_type_options', None)
+  all_vars_dict.pop('filing_type_map', None)
+  all_vars_dict.pop('party_type_options', None)
+  return json.dumps(all_vars_dict)
 
 class ProxyConnection:
   def __init__(self, url:str=None, api_key:str=None, credentials_code_block:str='tyler_login'):
@@ -64,13 +81,13 @@ class ProxyConnection:
       'Content-type': 'application/json',
       'Accept': 'application/json',
     }
-    self.proxy_client.headers['X-API-KEY'] = api_key 
+    self.proxy_client.headers['X-API-KEY'] = api_key
     self.verbose = False
     self.authed_user_id = None
     self.set_verbose_logging(True)
     self.credentials_code_block = credentials_code_block
-    
-  def _call_proxy(self, send_func):
+
+  def _call_proxy(self, send_func) -> ApiResponse:
     try:
       resp = send_func()
       if resp.status_code == 401 and self.credentials_code_block:
@@ -98,7 +115,7 @@ class ProxyConnection:
     self.verbose = turn_on
 
   @staticmethod
-  def user_visible_resp(resp):
+  def user_visible_resp(resp) -> ApiResponse:
     """By default, the responses from `requests` aren't pickable, which causes
         some issues in Docassemble if you are using this in interviews. This function
         takes the essentials of a response and puts in into a simple object.
@@ -129,10 +146,10 @@ class ProxyConnection:
     if tyler_email and tyler_password:
       auth_obj['tyler'] = {'username': tyler_email, 'password': tyler_password}
     try:
-      resp = self.proxy_client.post(self.base_url + 'adminusers/authenticate', 
+      resp = self.proxy_client.post(self.base_url + 'adminusers/authenticate',
         data=json.dumps(auth_obj))
       if resp.status_code == requests.codes.ok:
-        all_tokens = resp.json()
+        all_tokens = resp.json().get('tokens', {})
         for k, v in all_tokens.items():
           self.proxy_client.headers[k] = v
         # self.authed_user_id = data['userID']
@@ -145,8 +162,8 @@ class ProxyConnection:
 
   def register_user(self, person, registration_type:str, password:str=None, firm_name_or_id:str=None):
     """
-    registration_type needs to be INDIVIDUAL, FIRM_ADMINISTRATOR, or FIRM_ADMIN_NEW_MEMBER. 
-    If registration_type is INDIVIDUAL or FIRM_ADMINISTRATOR, you need a password. 
+    registration_type needs to be INDIVIDUAL, FIRM_ADMINISTRATOR, or FIRM_ADMIN_NEW_MEMBER.
+    If registration_type is INDIVIDUAL or FIRM_ADMINISTRATOR, you need a password.
     If it's FIRM_ADMINISTRATOR or FIRM_ADMIN_NEW_MEMBER, you need a firm_name_or_id
     """
     if hasattr(person, 'phone_number'):
@@ -191,7 +208,7 @@ class ProxyConnection:
     try:
       return bool(re.match(results.data.get('regularexpression'), password))
     except:
-      return None      
+      return None
 
   # Managing Firm Users
 
@@ -211,27 +228,27 @@ class ProxyConnection:
 
   def get_user_roles(self, id:str):
     send = lambda: self.proxy_client.get(self.base_url + f'adminusers/users/{id}/roles')
-    return self._call_proxy(send) 
+    return self._call_proxy(send)
 
   def add_user_roles(self, id:str, roles:List[dict]):
-    send = lambda: self.proxy_client.post(self.base_url + f'adminusers/users/{id}/roles', 
+    send = lambda: self.proxy_client.post(self.base_url + f'adminusers/users/{id}/roles',
       data=json.dumps(roles))
-    return self._call_proxy(send) 
+    return self._call_proxy(send)
 
   def remove_user_roles(self, id:str, roles:List[dict]):
     send = lambda: self.proxy_client.delete(self.base_url + f'adminusers/users/{id}/roles',
         data=json.dumps(roles))
-    return self._call_proxy(send) 
+    return self._call_proxy(send)
 
-  def change_password(self, id:str, email:str, new_password:str): 
+  def change_password(self, id:str, email:str, new_password:str):
     send = lambda: self.proxy_client.post(self.base_url + f'adminusers/users/{id}/password',
         data=json.dumps({'email': email, 'newPassword': new_password}))
-    return self._call_proxy(send) 
+    return self._call_proxy(send)
 
   def self_change_password(self, current_password:str, new_password:str):
     send = lambda: self.proxy_client.post(self.base_url + f'adminusers/user/password',
         data=json.dumps({'currentPassword': current_password, 'newPassword': new_password}))
-    return self._call_proxy(send) 
+    return self._call_proxy(send)
 
   def self_resend_activation_email(self, email: str):
     send = lambda: self.proxy_client.post(self.base_url + 'adminusers/user/resend-activation-email',
@@ -240,10 +257,10 @@ class ProxyConnection:
 
   def resend_activation_email(self, id:str):
     send = lambda: self.proxy_client.post(self.base_url + f'adminusers/users/{id}/resend-activation-email')
-    return self._call_proxy(send) 
+    return self._call_proxy(send)
 
   def update_notification_preferences(self, notif_preferences:List[dict]):
-    send = lambda: self.proxy_client.patch(self.base_url + f'adminusers/user/notification-preferences', 
+    send = lambda: self.proxy_client.patch(self.base_url + f'adminusers/user/notification-preferences',
         data=json.dumps(notif_preferences))
     return self._call_proxy(send)
 
@@ -255,12 +272,12 @@ class ProxyConnection:
     """AKA NotificationPreferencesList"""
     send = lambda: self.proxy_client.get(self.base_url + 'adminusers/notification-options')
     return self._call_proxy(send)
-  
+
   def self_update_user(self, email:str=None, first_name:str=None, middle_name:str=None, last_name:str=None):
     updated_user = {'email': email, 'firstName': first_name, 'middleName': middle_name, 'lastName': last_name}
     send = lambda: self.proxy_client.patch(self.base_url + f'adminusers/user', data=json.dumps(updated_user))
     return self._call_proxy(send)
-  
+
   def update_user(self, id:str, email:str=None, first_name:str=None, middle_name:str=None, last_name:str=None):
     updated_user = {'email': email, 'firstName': first_name, 'middleName': middle_name, 'lastName': last_name}
     send = lambda: self.proxy_client.patch(self.base_url + f'adminusers/users/{id}', data=json.dumps(updated_user))
@@ -289,20 +306,20 @@ class ProxyConnection:
   def update_firm(self, firm:Person):
     # firm is stateful
     update = serialize_person(firm)
-    
+
     send = lambda: self.proxy_client.patch(self.base_url + f'firmattorneyservice/firm', data=json.dumps(update))
-    return self._call_proxy(send) 
+    return self._call_proxy(send)
 
   # Managing Attorneys
   def get_attorney_list(self):
     send = lambda: self.proxy_client.get(self.base_url + f'firmattorneyservice/attorneys')
-    return self._call_proxy(send) 
+    return self._call_proxy(send)
 
   def get_attorney(self, attorney_id):
     send = lambda: self.proxy_client.get(self.base_url + f'firmattorneyservice/attorneys/{attorney_id}')
     return self._call_proxy(send)
 
-  def update_attorney(self, attorney_id, bar_number:str=None, 
+  def update_attorney(self, attorney_id, bar_number:str=None,
       first_name:str=None, middle_name:str=None, last_name:str=None):
     send = lambda: self.proxy_client.patch(self.base_url + f'firmattorneyservice/attorneys/{attorney_id}', 
         data=json.dumps({'barNumber': bar_number, 'firstName': first_name, 'middleName': middle_name, 'lastName': last_name}))
@@ -311,16 +328,16 @@ class ProxyConnection:
   def create_attorney(self, bar_number:str, first_name:str, middle_name:str=None, last_name:str=None):
     send = lambda: self.proxy_client.post(self.base_url + f'firmattorneyservice/attorneys', 
         data=json.dumps({'barNumber': bar_number, 'firstName': first_name, 'middleName': middle_name, 'lastName': last_name}))
-    return self._call_proxy(send) 
+    return self._call_proxy(send)
 
   def remove_attorney(self, attorney_id):
     send = lambda: self.proxy_client.delete(self.base_url + f'firmattorneyservice/attorneys/{attorney_id}')
-    return self._call_proxy(send) 
+    return self._call_proxy(send)
 
   # Managing Payment Accounts
   def get_payment_account_type_list(self):
     send = lambda: self.proxy_client.get(self.base_url + f'payments/types')
-    return self._call_proxy(send) 
+    return self._call_proxy(send)
 
   def get_payment_account_list(self):
     send = lambda: self.proxy_client.get(self.base_url + f'payments/payment-accounts')
@@ -373,8 +390,11 @@ class ProxyConnection:
     send = lambda: self.proxy_client.get(self.base_url + f'firmattorneyservice/service-contacts/{service_contact_id}')
     return self._call_proxy(send)
 
-  def update_service_contact(self, service_contact_id, service_contact:Individual, is_public:bool, is_in_master_list:bool, admin_copy:str=None):
+  def update_service_contact(self, service_contact_id, service_contact:Individual, is_public:bool=None, is_in_master_list:bool=None, admin_copy:str=None):
     service_contact_dict = serialize_person(service_contact)
+    service_contact_dict['isPublic'] = is_public
+    service_contact_dict['isInFirmMasterList'] = is_in_master_list
+    service_contact_dict['administrativeCopy'] = admin_copy
     send = lambda: self.proxy_client.patch(self.base_url + f'firmattorneyservice/service-contacts/{service_contact_id}', 
         data=json.dumps(service_contact_dict))
     return self._call_proxy(send)
@@ -415,11 +435,6 @@ class ProxyConnection:
     send = lambda: self.proxy_client.get(url)
     return self._call_proxy(send)
 
-  def get_case_service_history(self, court_id:str, case_tracking_id:str):
-    url = self.base_url + f"cases/courts/{court_id}/cases/{case_tracking_id}/service-information-history"
-    send = lambda: self.proxy_client.get(url)
-    return self._call_proxy(send)
-  
   def get_public_list(self):
     send = lambda: self.proxy_client.get(self.base_url + f'firmattorneyservice/service-contacts/public')
     return self._call_proxy(send)
@@ -440,11 +455,11 @@ class ProxyConnection:
     send = lambda: self.proxy_client.get(self.base_url + f'filingreview/jurisdictions/{jurisdiction}/courts')
     return self._call_proxy(send)
 
-  def get_filing_list(self, jurisdiction:str, court_id:str, start_date:DADateTime, end_date:DADateTime):
+  def get_filing_list(self, jurisdiction:str, court_id:str, user_id:str=None, start_date:DADateTime=None, end_date:DADateTime=None):
     params = {
-      "user_id": None,
-      "start_date": as_datetime(start_date).format("yyyy-MM-dd"),
-      "end_date": as_datetime(end_date).format("yyyy-MM-dd")
+      "user_id": user_id,
+      "start_date": as_datetime(start_date).format("yyyy-MM-dd") if start_date else None,
+      "end_date": as_datetime(end_date).format("yyyy-MM-dd") if end_date else None
     }
     send = lambda: self.proxy_client.get(self.base_url + f'filingreview/jurisdictions/{jurisdiction}/courts/{court_id}/filings', params=params)
     return self._call_proxy(send) 
@@ -454,7 +469,7 @@ class ProxyConnection:
     return self._call_proxy(send) 
 
   def get_policy(self, jurisdiction:str, court_id:str):
-    send = lambda: self.proxy_client.get(self.base_url + f'filingreview/jurisdicitions/{jurisdiction}/courts/{court_id}/policy')
+    send = lambda: self.proxy_client.get(self.base_url + f'filingreview/jurisdictions/{jurisdiction}/courts/{court_id}/policy')
     return self._call_proxy(send)
 
   def get_filing_status(self, jurisdiction:str, court_id:str, filing_id:str):
@@ -465,42 +480,56 @@ class ProxyConnection:
     send = lambda: self.proxy_client.delete(self.base_url + f'filingreview/jurisdictions/{jurisdiction}/courts/{court_id}/filings/{filing_id}')
     return self._call_proxy(send) 
 
-  def check_filing(self, jurisdiction:str, court_id:str, court_bundle:ALDocumentBundle):
-    _recursive_give_data_url(court_bundle)
-    all_vars = json.dumps(all_variables())
-    log(f'headers: {self.proxy_client.headers}')
+  def check_filing(self, jurisdiction:str, court_id:str, court_bundle:ALDocumentBundle=None):
+    all_vars = _get_all_vars(court_bundle) 
     send = lambda: self.proxy_client.get(self.base_url + f'filingreview/jurisdictions/{jurisdiction}/courts/{court_id}/filing/check', data=all_vars)
     return self._call_proxy(send)
 
-  def file_for_review(self, jurisdiction: str, court_id:str, court_bundle:ALDocumentBundle):
-    _recursive_give_data_url(court_bundle)
-    all_vars_obj = all_variables()
-    all_vars = json.dumps(all_vars_obj)
+  def file_for_review(self, jurisdiction: str, court_id:str, court_bundle:ALDocumentBundle=None):
+    all_vars = _get_all_vars(court_bundle) 
     send = lambda: self.proxy_client.post(self.base_url + f'filingreview/jurisdictions/{jurisdiction}/courts/{court_id}/filings',
           data=all_vars)
     return self._call_proxy(send)
+
+  def get_service_types(self, jurisdiction:str, court_id:str, court_bundle:ALDocumentBundle=None):
+    """Checks the court info: if it has conditional service types, call a special API with all filing info so far to get service types"""
+    court_info = self.get_court(court_id)
+    if court_info.data.get('hasconditionalservicetypes'):
+      all_vars = _get_all_vars(court_bundle) 
+      send = lambda: self.proxy_client.get(self.base_url + f'filingreview/jurisdictions/{jurisdiction}/courts/{court_id}/filing/servicetypes',
+        data=all_vars)
+      return self._call_proxy(send)
+    else:
+      return self.get_service_type_codes(court_id)
+      
+  # TODO(brycew): rethink service API
+  #def serve(self, jurisdiction:str, court_id:str, court_bundle:ALDocumentBundle=None):
+  #  all_vars = _get_all_vars(court_bundle) 
+  #  send = lambda: self.proxy_client.post(self.base_url + f'filingreview/jurisdictions/{jurisdiction}/courts/{court_id}/filing/serve',
+  #                                        data=all_vars)
+  #  return self._call_proxy(send)
   
-  def calculate_filing_fees(self, jurisdiction:str, court_id:str, court_bundle:ALDocumentBundle):
-    _recursive_give_data_url(court_bundle)
-    all_vars_obj = all_variables()
-    all_vars = json.dumps(all_vars_obj)
+  def calculate_filing_fees(self, jurisdiction:str, court_id:str, court_bundle:ALDocumentBundle=None):
+    all_vars = _get_all_vars(court_bundle)
     send = lambda: self.proxy_client.post(self.base_url + f'filingreview/jurisdictions/{jurisdiction}/courts/{court_id}/filing/fees',
           data=all_vars)
-    return self._call_proxy(send)    
+    return self._call_proxy(send)
 
   def get_return_date(self, court_id:str, req_return_date, court_bundle:ALDocumentBundle):
-    _recursive_give_data_url(court_bundle)
+    _give_data_url(court_bundle)
     all_vars_obj = all_variables()
     all_vars_obj['return_date'] = req_return_date.isoformat()
     send = lambda: self.proxy_client.post(self.base_url + f'scheduling/courts/{court_id}/return_date', data=json.dumps(all_vars_obj))
     return self._call_proxy(send)
 
-  def reserve_court_date(self, court_id:str, doc_id:str, 
+  def reserve_court_date(self, court_id:str, doc_id:str,
       range_after:datetime=None, range_before:datetime=None, estimated_duration=None):
-    if range_after is not None:
+    if range_after is not None and not isinstance(range_after, str):
       range_after = range_after.isoformat()
-    if range_before is not None:
+    log(f'after: {range_after}')
+    if range_before is not None and not isinstance(range_before, str):
       range_before = range_before.isoformat()
+    log(f'before: {range_before}')
     if estimated_duration is not None:
       estimated_duration = int(estimated_duration) * 60 * 60
     send = lambda: self.proxy_client.post(self.base_url + f'scheduling/courts/{court_id}/reserve_date',
@@ -517,7 +546,12 @@ class ProxyConnection:
   
   def _get_cases(self, court_id:str, person_name:dict=None, business_name:str=None, docket_id:str=None) -> ApiResponse:
     send = lambda: self.proxy_client.get(self.base_url + f'cases/courts/{court_id}/cases', 
-        data=json.dumps({'person_name': person_name, 'business_name': business_name, 'docket_id': docket_id}))
+        params={
+          'first_name': person_name.get('first') if person_name is not None else None,
+          'middle_name': person_name.get('middle') if person_name is not None else None,
+          'last_name': person_name.get('last') if person_name is not None else None,
+          'business_name': business_name,
+          'docket_id': docket_id})
     return self._call_proxy(send)
 
   def get_case(self, court_id:str, case_id:str):
@@ -532,20 +566,30 @@ class ProxyConnection:
     send = lambda: self.proxy_client.get(self.base_url + f'cases/courts/{court_id}/service-contacts/{service_contact_id}/cases')
     return self._call_proxy(send)
 
-  def get_service_information(self, court_id:str, case_id:str):
-    send = lambda: self.proxy_client.get(self.base_url + f'cases/courts/{court_id}/cases/{case_id}/service-information')
+  def get_service_information(self, court_id:str, case_tracking_id:str):
+    send = lambda: self.proxy_client.get(
+        self.base_url + f'cases/courts/{court_id}/cases/{case_tracking_id}/service-information')
     return self._call_proxy(send)
 
-  def get_service_information_history(self, court_id:str, case_id:str):
-    send = lambda: self.proxy_client.get(self.base_url + f'cases/courts/{court_id}/cases/{case_id}/service-information-history')
+  def get_service_information_history(self, court_id:str, case_tracking_id:str):
+    send = lambda: self.proxy_client.get(
+        self.base_url + f'cases/courts/{court_id}/cases/{case_tracking_id}/service-information-history')
     return self._call_proxy(send)
   
-  def get_case_categories(self, court_id:str, filable_only:bool=False):
-    send = lambda: self.proxy_client.get(self.base_url + f'codes/courts/{court_id}/categories?filable_only={filable_only}')
+  def get_case_categories(self, court_id:str, fileable_only:bool=False, timing:str=None):
+    params = {
+      "fileable_only": fileable_only,
+      "timing": timing 
+    }
+    send = lambda: self.proxy_client.get(self.base_url + f'codes/courts/{court_id}/categories', params=params) 
     return self._call_proxy(send)
   
   def get_case_types(self, court_id:str, case_category:str, timing:str=None):
-    send = lambda: self.proxy_client.get(self.base_url + f'codes/courts/{court_id}/case_types?category_id={case_category}&timing={timing}')
+    params = {
+      'category_id': case_category,
+      'timing': timing
+    }
+    send = lambda: self.proxy_client.get(self.base_url + f'codes/courts/{court_id}/case_types', params=params) 
     return self._call_proxy(send)
   
   def get_case_subtypes(self, court_id:str, case_type:str):
@@ -553,10 +597,15 @@ class ProxyConnection:
     return self._call_proxy(send)
   
   def get_filing_types(self, court_id:str, case_category:str, case_type:str, initial:bool):
-    send = lambda: self.proxy_client.get(self.base_url + f'codes/courts/{court_id}/filing_types?category_id={case_category}&type_id={case_type}&initial={initial}')
+    params = {
+      'category_id': case_category,
+      'type_id': case_type,
+      'initial': initial
+    }
+    send = lambda: self.proxy_client.get(self.base_url + f'codes/courts/{court_id}/filing_types', params=params)
     return self._call_proxy(send)
   
-  def get_service_types(self, court_id:str): 
+  def get_service_type_codes(self, court_id:str): 
     send = lambda: self.proxy_client.get(self.base_url + f'codes/courts/{court_id}/service_types')
     return self._call_proxy(send)
   
